from typing import List, Type, NamedTuple
from dataclasses import dataclass

import pytest

from blocks.compat import HAS_PYDANTIC

from blocks.db.next.types import Table


class RowNT(NamedTuple):
    id: int


class TableNT(Table):
    rows: List[RowNT]


@dataclass
class RowDC:
    id: int


class TableDC(Table):
    rows: List[RowDC]


types_mapping = {
    RowNT: TableNT,
    RowDC: TableDC,
}


if HAS_PYDANTIC:
    from pydantic import BaseModel

    class RowP(BaseModel):
        id: int


    class TableP(Table):
        rows: List[RowP]

    types_mapping.update({RowP: TableP})


@pytest.mark.parametrize('row,table', [(row, table) for row, table in types_mapping.items()])
def test_table(row: Type[object], table: Type[Table]):
    t = table(rows=[])
    assert t.columns == ['id']
    assert t.values == []

    t.rows.append(row(id=1))
    assert t.columns == ['id']
    assert t.values == [(1,)]



from __future__ import annotations

import time
from collections import OrderedDict, defaultdict
from types import TracebackType
from typing import Optional, Type, List, Dict

from blocks.types.base import Processor, Event
from blocks.types.metrics import EventTime, AggregatedMetric


class _ProcTimer:

    def __enter__(self) -> _ProcTimer:
        self.start = time.perf_counter()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType]
    ) -> None:
        self.end = time.perf_counter()


class MetricCollector(object):
    """
    Collecting information about processor operation


    - The number of events passed through the processor in N time (1 minute by default)
    - Maximum, minimum and average operating time of the processor with 1 event in N time (1 minute by default)

    Example::

      >>> from blocks import App, source, processor
      >>> from blocks.types.metrics import AggregatedMetric

      >>> @processor
      >>> def printer(e: AggregatedMetric) -> None:
      ...       print(e)

      >>> blocks  = (..., printer())
      >>> App(blocks, collect_metric=True).run()
    """

    def __init__(self, metric_time_interval: int) -> None:
        """
        Init metric collector instance

        :param metric_time_interval:    Time interval for metric aggregation.
        """

        self._tc = metric_time_interval
        self._collected_data: Dict[Processor, Dict[Event, List[EventTime]]] = OrderedDict()

        self.timer = _ProcTimer()

    def collect(self, processor_type: Processor, event_time: EventTime) -> None:
        if processor_type not in self._collected_data:
            self._collected_data[processor_type] = defaultdict(list)
        self._collected_data[processor_type][event_time.event].append(event_time)

    def get_aggregate_events(self) -> List[AggregatedMetric]:
        agg_metrics: List[AggregatedMetric] = []
        to_delete = []
        for processor, events in self._collected_data.items():
            for event, times in events.items():
                interval = times[-1].end - times[0].start
                if interval < self._tc:
                    continue
                intervals = [t.interval for t in times]
                agg_metrics.append(
                    AggregatedMetric(
                        processor=processor,
                        interval=round(interval, 2),
                        type_event=event,
                        count_events=len(times),
                        max_processing=max(intervals),
                        min_processing=min(intervals),
                        avg_processing=round(sum(intervals) / len(intervals), 2),
                    )
                )
                to_delete.append((processor, event))
        for pr, ev in to_delete:
            del self._collected_data[pr][ev]

        return agg_metrics



from __future__ import annotations

import inspect
from typing import Any, List, Union, Callable

from blocks import Event, Source
from blocks.logger import logger
from blocks.db.types import Row, Query, Table

from pyhive.hive import Connection


def _get_rows_or_table(conn: Connection, query: Query, table_name: str) -> Union[List[Row], Table]:
    events = []
    if not (issubclass(query.codec, Row) or issubclass(query.codec, Table)):
        raise TypeError('Not valid codec type: {0}'.format(type(query.codec)))
    row_codec = query.codec if issubclass(query.codec, Row) else query.codec.type()
    with conn.cursor() as cur:
        logger.info(f'Executing query: {query.text}')
        cur.execute(query.text)
        logger.info('Validating')

        # Each column in description returns Tuple where first value is column name.
        keys = [col[0] for col in cur.description]
        # Column name contains table name - remove it for validation
        if keys[0].startswith('{0}.'.format(table_name)):
            keys = ['.'.join(col.split('.')[1:]) for col in keys]
        fields = inspect.signature(row_codec).parameters
        fields_match = set(keys) == set(fields)
        if fields_match is False:
            logger.warning("Looks like query doesn't exactly matches model. Please, check.")
        for row in cur:
            row_dict = {key: value for key, value in zip(keys, row)}
            if fields_match is False:
                row_dict = {key: row_dict[key] for key in fields}
            events.append(row_codec(**row_dict))
    if issubclass(query.codec, Row):
        return events
    elif issubclass(query.codec, Table):
        return query.codec(rows=events)
    raise TypeError('Not valid codec type: {0}'.format(type(query.codec)))


class HiveReader(Source):
    def __init__(self, queries: List[Query], conn: Callable[[], Connection], table_name: str) -> None:
        self._conn_factory = conn
        self._table_name = table_name
        self._queries = queries
        self.patch_annotations({query.codec for query in queries})

    def __call__(self) -> List[Event]:
        conn = self._conn_factory()
        with conn:
            rows_or_tables = self._run_queries(conn)
        return rows_or_tables

    def _run_queries(self, conn: Any) -> List[Event]:
        rows_or_tables: List[Event] = []
        for query in self._queries:
            row_or_table = _get_rows_or_table(conn, query, self._table_name)
            if isinstance(row_or_table, list):
                rows_or_tables.extend(row_or_table)
            else:
                rows_or_tables.append(row_or_table)
        return rows_or_tables



from __future__ import annotations

from typing import List, Type, Tuple
from dataclasses import dataclass

from blocks import Event


# ToDo (tribunsky.kir): make Table generic which is parametrized via event for rows.
@dataclass
class Table(Event):
    """
    Base class for database table description.
    Performs runtime type checks on every received row
    from query and drops invalid rows and then groups
    rows into single table. Database source will return
    all rows grouped in list as a single event in this case.
    Example::

      >>> from typing import Optional
      >>> from blocks.db import Row, Table
      >>>
      >>> class SomeTableRow(Row):
      ...     field1: Optional[int]
      ...     field2: str
      >>>
      >>> class SomeTable(Table):
      ...     rows: List[SomeTableRow]
      >>>
      >>> query = Query('SELECT * FROM some_table', SomeTable)
    """
    rows: List[Event]

    @property
    def columns(self) -> List[str]:
        # ToDo (tribunsky.kir): stupid, but will work for simple cases
        tp = self.type()
        return list(tp.__annotations__)

    @property
    def values(self) -> List[Tuple]:
        if self.rows:
            values = []
            fields = self.columns
            for row in self.rows:
                tpl = tuple(getattr(row, col) for col in fields)
                values.append(tpl)
            return values
        return []

    @classmethod
    def type(cls) -> Type[Event]:
        # pydantic:
        # row_codec = query.codec.__annotations__['rows'].__args__[0]
        # dataclass:
        # return cls.rows.__args__[0]
        [tp] = cls.__annotations__['rows'].__args__
        return tp



from dataclasses import dataclass

from blocks.types import Processor, Event


@dataclass
class EventTime:
    event: Event
    start: float
    end: float

    @property
    def interval(self) -> float:
        return self.end - self.start


@dataclass
class AggregatedMetric:
    processor: Processor
    interval: float
    type_event: Event
    count_events: int
    max_processing: float
    min_processing: float
    avg_processing: float

    def __str__(self) -> str:
        return '[{0} - {1}] Processed {2} events for {3}. MIN: {4}, MAX: {5}, AVG: {6}'.format(
            self.processor,
            self.type_event,
            self.count_events,
            self.interval,
            self.min_processing,
            self.max_processing,
            self.avg_processing
        )



import time
from typing import List, Optional

import pytest

try:
    import schedule
except ImportError:
    pytest.skip("skipping schedule-only tests", allow_module_level=True)

from blocks import Event
from blocks.sources.schedule import Scheduler


class ScheduledTrigger(Event):
    ...


def init_source(*, skip_when_busy: Optional[bool] = None) -> Scheduler:
    if skip_when_busy is None:
        source = Scheduler(ScheduledTrigger).every(1).seconds
    else:
        source = Scheduler(ScheduledTrigger, skip_when_busy=skip_when_busy).every(1).seconds
    source()
    return source


TEST_CASES = [
    # Default behaviour
    (None, 5.1, 0.01, 5),
    # Long work in main loop
    (None, 4.1, 2.0, 2),
    # Long work in main loop (2)
    (None, 5.1, 5.0, 1),
    # Should behave same as default
    (False, 5.1, 0.01, 5),
    # Should not behave same as long work as we emit late events
    (False, 5.1, 5.0, 5),
]


@pytest.mark.parametrize("skip_when_busy,duration,delay,answer", TEST_CASES)
def test_scheduler(skip_when_busy: bool, duration: float, delay: float, answer: int) -> None:
    source = init_source(skip_when_busy=skip_when_busy)
    events = []
    t0 = time.perf_counter()
    while time.perf_counter() - t0 < duration:
        new_events = source()
        assert isinstance(new_events, list)
        events.extend(new_events)
        time.sleep(delay)
    assert len(events) == answer



import asyncio
from dataclasses import dataclass

from blocks import App, AsyncSource, AsyncProcessor, async_source, async_processor


@dataclass
class E:
    x: int


@dataclass
class E2:
    y: int


@async_source
async def gen() -> E:
    await asyncio.sleep(0.1)
    return E(1)


@async_processor
async def printer(e: E) -> E2:
    print('async_processor', e)
    return E2(e.x)


class Gen(AsyncSource):
    async def __call__(self) -> E:
        await asyncio.sleep(0.2)
        return E(2)


class Printer(AsyncProcessor):
    async def __call__(self, event: E) -> E2:
        print('AsyncProcessor', event)
        return E2(event.x)


class Receiver(AsyncProcessor):
    def __init__(self) -> None:
        self.events = []

    async def __call__(self, event: E2) -> None:
        self.events.append(event)


def test_smoke_async_app() -> None:
    loop = asyncio.get_event_loop()
    receiver = Receiver()
    app = App(blocks=(gen(), Gen(), printer(), Printer(), receiver))
    loop.run_until_complete(app.run_async(once=True))
    assert len(receiver.events) == 4



from blocks.types.multiprocess import ParallelEvent
from blocks.types.base import (
    Block,
    Event,
    Source,
    AnySource,
    Processor,
    AsyncSource,
    AnyProcessor,
    EventOrEvents,
    AsyncProcessor,
    TypeOfProcessor,
)



import sys
from typing import Union, Optional
from dataclasses import dataclass

from blocks import Event, Graph, processor


@dataclass
class MyEvent(Event):
    x: int


@dataclass
class MyOtherEvent(MyEvent):
    y: float


@processor
def printer(event: MyEvent) -> MyOtherEvent:
    print(event)
    return MyOtherEvent(event.x, 1.0)


def test_event_not_in_graph() -> None:
    graph = Graph()

    assert MyEvent not in graph.outputs
    assert MyOtherEvent not in graph.outputs


def test_event_in_graph() -> None:
    graph = Graph()
    graph.add_block(printer())

    assert MyEvent not in graph.outputs
    assert MyOtherEvent in graph.outputs


if sys.version_info >= (3, 9):
    @processor
    def optional_list_processor(event: MyEvent) -> Optional[list[Union[MyEvent, MyOtherEvent]]]:
        return [event, MyOtherEvent(event.x, 1.0)]


    def test_event_in_graph_specific_annotation() -> None:
        graph = Graph()
        graph.add_block(optional_list_processor())

        assert MyEvent in graph.outputs
        assert MyOtherEvent in graph.outputs



import os
import sqlite3
from typing import Callable, Generator
from sqlite3 import Connection

import pytest


@pytest.fixture
def connect() -> Callable[[], Connection]:
    def connection_factory() -> Connection:
        con = sqlite3.connect('example.db')
        with con:
            con.execute('''CREATE TABLE IF NOT EXISTS test_table (id, name, text)''')
        return con

    return connection_factory


# FixMe (tribunsky.kir): potentially dangerous, use smth like mkstemp
@pytest.fixture(autouse=True)
def run_around_tests() -> Generator[None, None, None]:
    file_name = 'example.db'
    if os.path.exists(file_name):
        raise
    try:
        yield
    finally:
        if os.path.exists(file_name):
            os.remove(file_name)



from typing import Dict, List, Tuple, Union, Optional, NamedTuple
from dataclasses import dataclass

import pytest

try:
    from redis import Redis
except ImportError:
    pytest.skip("skipping redis-only tests", allow_module_level=True)
from pytest import approx
from pydantic import BaseModel

from blocks.redis import InputStream, RedisConsumer


class RedisStub(Redis):
    def xread(
        self,
        streams: Dict[bytes, bytes],
        count: Optional[int] = None,
        block: Optional[int] = None,
    ) -> List[List[Union[bytes, List[Tuple[bytes, Dict[bytes, bytes]]]]]]:
        return [[b'test', [(b'1667229601456-0', {b'id': b'181489903', b'value': b'0.0023', b'flag': b'false'})]]]


class SignalNT(NamedTuple):
    id: int
    value: float
    flag: bool


class SignalO(object):
    id: int
    value: float
    flag: bool


class SignalOI(object):
    def __init__(self, id: int, value: float, flag: bool) -> None:
        self.id = id
        self.value = value
        self.flag = flag


@dataclass
class SignalDC:
    id: int
    value: float
    flag: bool


class SignalP(BaseModel):
    id: int
    value: float
    flag: bool


# ToDo (tribunsky.kir): currently only pydantic-based events work out-of-the box.
#                       for enlightened events decoding is necessary or client with decode_responses=True may help.
test_types = [
    # SignalNT,
    # SignalO,
    # SignalOI,
    # SignalDC,
    SignalP,
]


@pytest.fixture
def client() -> Redis:
    return RedisStub()


@pytest.mark.parametrize('cls', test_types)
def test_smoke_event_creation(cls, client) -> None:
    streams = [InputStream(name='test', event=cls)]
    source = RedisConsumer(client, streams, read_timeout=100)

    events = source()
    assert events
    for event in events:
        assert isinstance(event, cls)
        assert event.id == 181489903
        assert event.value == approx(0.0023)
        assert event.flag is False



import os
from typing import Generator

import pytest


@pytest.fixture(autouse=True)
def run_around_tests() -> Generator[None, None, None]:
    files = ['GRAPH.png', 'DAG.svg']
    for file_name in files:
        if os.path.exists(file_name):
            raise
    try:
        yield
    finally:
        for file_name in files:
            if os.path.exists(file_name):
                os.remove(file_name)



import logging
from typing import Any, List, NamedTuple
from dataclasses import dataclass

import pytest

from blocks.types import Event, Processor
from blocks.decorators import source, processor
from blocks.validation import validate_blocks, validate_annotations
from blocks.annotations import AnnotationError


class ListHandler(logging.Handler):
    def __init__(self, *args: Any, message_list: List[str], **kwargs: Any) -> None:
        logging.Handler.__init__(self, *args, **kwargs)
        self.message_list = message_list

    def emit(self, record: logging.LogRecord) -> None:
        self.message_list.append(self.format(record).rstrip('\n'))


class E1(Event):
    def __init__(self, x: int) -> None:
        self.x = x


class E2(NamedTuple):
    x: int


@dataclass
class E3:
    x: int


@processor
def p1(e: E1) -> E2:
    return E2(x=1)


@processor
def p2(e: E1):
    return E2(x=1)


@processor
def p3(e):
    return E2(x=1)


class P4(Processor):
    def __call__(self, event: E1):
        return E2(x=1)


class P5(Processor):
    def __call__(self, e1: E1, e2: E2) -> E3:
        return E3(x=1)


class P6(Processor):
    def __call__(self, event: E1) -> E2:
        return E2(x=2)


def test_valid_annotations() -> None:
    validate_annotations(p1())
    validate_annotations(P6())


def test_invalid_annotations() -> None:
    with pytest.raises(AnnotationError):
        validate_annotations(p2())
    with pytest.raises(AnnotationError):
        validate_annotations(p3())
    with pytest.raises(AnnotationError):
        validate_annotations(P4())
    with pytest.raises(AnnotationError):
        validate_annotations(P5())


@source
def g1() -> E1:
    return E1(x=1)


@processor
def d1(e: E2) -> None:
    ...


def test_invalid_blocks() -> None:
    logger = logging.getLogger('blocks')
    messages = []
    logger.addHandler(ListHandler(message_list=messages))

    validate_blocks([p1()])
    assert len(messages) == 3
    assert 'E2' in messages[0] and "isn't processed" in messages[0]
    assert 'p1' in messages[1] and 'There is no events generated for' in messages[1]
    assert 'p1' in messages[2] and 'There is no receivers for generated events' in messages[2]
    messages.clear()

    validate_blocks([g1(), p1()])
    assert len(messages) == 2
    assert 'E2' in messages[0] and "isn't processed" in messages[0]
    assert 'p1' in messages[1] and 'There is no receivers for generated events' in messages[1]


def test_valid_blocks() -> None:
    logger = logging.getLogger('blocks')
    messages = []
    logger.addHandler(ListHandler(message_list=messages))

    validate_blocks([g1(), p1(), d1()])
    assert not messages



import os
import time
from dataclasses import dataclass
from typing import List, Optional, Tuple

import pytest

from blocks import App, Event, parallel_processor, Processor, source

from blocks.types.multiprocess import dill

if dill is None:
    pytest.skip("skipping dill-only tests", allow_module_level=True)


@dataclass
class InputEvent(Event):
    value: int


@dataclass
class OutPutEvent(Event):
    value: int
    pid: int
    process: int


@dataclass
class TerminalEvent(Event):
    ...


@source
def generator() -> InputEvent:
    return InputEvent(42)


@parallel_processor
def parallel_process_1(e: InputEvent) -> OutPutEvent:
    time.sleep(3)
    return OutPutEvent(value=e.value * 2, pid=os.getpid(), process=1)


@parallel_processor
def parallel_process_2(e: InputEvent) -> OutPutEvent:
    time.sleep(3)
    return OutPutEvent(value=e.value * 4, pid=os.getpid(), process=2)


@parallel_processor
def parallel_process_3(e: InputEvent) -> OutPutEvent:
    time.sleep(3)
    return OutPutEvent(value=e.value * 6, pid=os.getpid(), process=3)


class PostProcess(Processor):
    def __init__(self) -> None:
        self.out_events: List[OutPutEvent] = []

    def __call__(self, e: OutPutEvent) -> Optional[TerminalEvent]:
        if len(self.out_events) >= 3:
            return TerminalEvent()
        self.out_events.append(e)


@pytest.mark.parametrize("blocks, count", (
        ((generator(), parallel_process_1(), PostProcess()), 1),
        ((generator(), parallel_process_1(), parallel_process_2(), PostProcess()), 2),
        ((generator(), parallel_process_1(), parallel_process_2(), parallel_process_3(), PostProcess()), 3)
))
def test_count_of_parallel_processor(blocks: Tuple, count: int) -> None:
    assert App(blocks=blocks)._graph.count_of_parallel_tasks == count


def test_run_parallel_processor() -> None:
    blocks = (generator(), parallel_process_1(), parallel_process_2(), parallel_process_3(), PostProcess())
    App(blocks=blocks, terminal_event=TerminalEvent).run()
    for event in blocks[-1].out_events:
        assert event.pid != os.getpid()
        assert event.value == event.process * 2 * 42



from typing import List, Union, NamedTuple
from dataclasses import dataclass

from blocks.types import Event, Processor
from blocks.decorators import source, processor
from blocks.annotations import get_input_events_type, get_output_events_type


class E1(Event):
    def __init__(self, x: int) -> None:
        self.x = x


class E2(NamedTuple):
    x: int


@dataclass
class E3:
    x: int


@processor
def proс0(event: E1) -> List[Union[E1, E2, E3]]:
    return [E2(x=event.x)]


class ClassProc0(Processor):
    def __call__(self, event: E1) -> List[Union[E1, E2, E3]]:
        return [E2(x=event.x)]


class ClassProc1(Processor):
    def __call__(self, event: Union[E1, E2, E3]) -> List[E2]:
        return [E2(x=event.x)]


@processor
def proc1(event: E1) -> List[E2]:
    return [E2(x=event.x)]


@processor
def proc2(event: E2) -> E3:
    return E3(x=event.x)


@processor
def proc3(event: Union[E1, E2, E3]) -> E3:
    return E3(x=event.x)


@processor
def printer(event: E3) -> None:
    print(event)


@source
def generator() -> E2:
    return E2(x=1)


def test_input_union_annotation_flatten() -> None:
    events = get_input_events_type(proc3())

    assert len(events) == 3
    assert events[0] is E1
    assert events[1] is E2
    assert events[2] is E3


def test_output_union_annotation_flatten() -> None:
    events = get_output_events_type(proс0())

    assert len(events) == 3
    assert events[0] is E1
    assert events[1] is E2
    assert events[2] is E3


def test_input_class_processor_union_annotation_flatten() -> None:
    events = get_input_events_type(ClassProc1())

    assert len(events) == 3
    assert events[0] is E1
    assert events[1] is E2
    assert events[2] is E3


def test_output_class_processor_union_annotation_flatten() -> None:
    events = get_output_events_type(ClassProc0())

    assert len(events) == 3
    assert events[0] is E1
    assert events[1] is E2
    assert events[2] is E3


def test_input_annotations_parse() -> None:
    events = get_input_events_type(proc1())

    assert len(events) == 1
    assert events[0] is E1


def test_output_annotations_parse() -> None:
    events = get_output_events_type(proc1())

    assert len(events) == 1
    assert events[0] is E2


def test_input_annotations_none() -> None:
    events = get_input_events_type(generator())

    assert len(events) == 0


def test_output_annotations_none() -> None:
    events = get_output_events_type(printer())

    assert len(events) == 0



import time
import array
import random
from typing import List, Union, Optional
from dataclasses import dataclass

from blocks import App, Event, Processor, source
from blocks.processors import Batcher, TimeoutedBatcher


@dataclass
class InputEvent(Event):
    camera_no: int
    timestamp: str
    image: array.array


@dataclass
class InputEventProcessed(Event):
    camera_no: int
    timestamp: str
    image: array.array


@dataclass
class ImagesBatch(Event):
    data: List[InputEvent]


@dataclass
class TriggerEvent(Event):
    ...


@dataclass
class InternalTriggerEvent(Event):
    ...


@dataclass
class TerminalEvent(Event):
    ...


@source
def generator() -> InputEvent:
    time.sleep(0.1)
    return InputEvent(random.randint(1, 9), "1668671381190", array.array('d', [2.4, 14.8, 7, 6]))


class PreProcessor(Processor):
    def __init__(self) -> None:
        self.trigger_enabled = False
        self.count = 0

    def __call__(
        self,
        e: Union[InputEvent, InternalTriggerEvent],
    ) -> Optional[Union[InputEventProcessed, TriggerEvent]]:
        if self.trigger_enabled:
            if self.count > 5:
                self.trigger_enabled = False
                self.count = 0
                return TriggerEvent()
            else:
                self.count += 1
        if isinstance(e, InputEvent):
            return InputEventProcessed(e.camera_no, e.timestamp, e.image)
        if isinstance(e, InternalTriggerEvent):
            self.trigger_enabled = True
        return None


class PostProcessor(Processor):
    def __init__(self) -> None:
        self.out_events: List[ImagesBatch] = []

    def __call__(self, e: ImagesBatch) -> Optional[TerminalEvent]:
        if len(self.out_events) >= 4:
            print('terminate')
            return TerminalEvent()
        self.out_events.append(e)
        return None


class TimeoutPostProcessor(Processor):
    def __init__(self) -> None:
        self.out_events: List[ImagesBatch] = []

    def __call__(self, e: ImagesBatch) -> Optional[Union[TerminalEvent, InternalTriggerEvent]]:
        self.out_events.append(e)
        if len(self.out_events) >= 4:
            return TerminalEvent()
        if len(self.out_events) == 2:
            return InternalTriggerEvent()
        return None


def test_run_batcher() -> None:
    blocks = (generator(), Batcher(InputEvent, ImagesBatch, 5), PostProcessor())
    App(blocks=blocks, terminal_event=TerminalEvent).run()


def test_run_timeout_batcher() -> None:
    blocks = (
        generator(),
        PreProcessor(),
        TimeoutedBatcher(InputEventProcessed, TriggerEvent, ImagesBatch, 10, 0.01),
        TimeoutPostProcessor(),
    )
    App(blocks=blocks, terminal_event=TerminalEvent).run()



import contextlib
from functools import wraps
from inspect import getfullargspec
from typing import Any, Callable, Dict, get_type_hints, Tuple, Type, Iterator, Optional

from blocks.types.ioc import Dependency


class Depends:
    def __init__(self, dependency: Any = None):
        self.dependency = dependency


def static_dependency(value: Any) -> Callable[[], Any]:
    @contextlib.contextmanager
    def manager() -> Iterator[Any]:
        yield value

    return manager


class Container:
    """
      >>> import contextlib
      >>> from dataclasses import dataclass
      >>> from typing import Iterator

      >>> from blocks.types.ioc import Dependency
      >>> from blocks.ioc import Container, Depends, static_dependency

      >>> def some_provider() -> None:
      >>>    raise NotImplementedError

      >>> def get_some() -> Dependency:
      >>>    @contextlib.contextmanager
      >>>    def manager() -> Iterator[int]:
      >>>        yield 42
      >>>    return manager

      >>> @dataclass
      >>> class Foo:
      >>>     value: str

      >>> c = Container(overrides={
      >>>         some_provider: get_some(),
      >>>         Foo: static_dependency(Foo(value="bar"))
      >>>     }
      >>> )

      >>> @c.inject
      >>> def show_info(some: int = Depends(some_provider), foo: Foo = Depends()) -> None:
      >>>     print(some)
      >>>     print(foo)

      >>> show_info()
    """
    __slots__ = ("_overrides",)

    def __init__(self, overrides: Dict[Callable, Dependency]):
        self._overrides = overrides

    def _normalize_dependecy(self, type_: Optional[Type], depends: Depends) -> Any:
        if depends.dependency is None:
            return type_
        return depends.dependency

    def _get_dependencies(self, func: Callable) -> Dict[str, Callable]:
        hints = get_type_hints(func)
        argspec = getfullargspec(func)
        defaults: Tuple[Depends, ...] = argspec.defaults or tuple()
        kwonlydefaults = argspec.kwonlydefaults or {}
        positional = {
            arg: self._normalize_dependecy(hints.get(arg), default)
            for arg, default in zip(argspec.args[::-1], defaults[::-1])
            if isinstance(default, Depends)
        }
        keyword = {
            arg: self._normalize_dependecy(hints.get(arg), default)
            for arg, default in kwonlydefaults.items()
            if isinstance(default, Depends)
        }
        return {
            **positional,
            **keyword,
        }

    def inject(self, handler: Callable) -> Callable:
        required_dependencies = self._get_dependencies(handler)

        @wraps(handler)
        def wrapper(*args: Tuple, **kwargs: Dict) -> Any:
            stack = contextlib.ExitStack()
            dependencies = {}
            for name, dependency in required_dependencies.items():
                dependency_context = self._overrides[dependency]()
                dependencies[name] = stack.enter_context(dependency_context)
            try:
                kwargs.update(dependencies)
                return handler(*args, **kwargs)
            finally:
                stack.close()

        return wrapper



import inspect
from typing import Any, List, Union, Callable
from sqlite3 import Connection

from blocks import Event, Source
from blocks.logger import logger
from blocks.db.next.sql import Query
from blocks.db.next.types import Table


def _get_rows_or_table(conn: Connection, query: Query) -> Union[List[Event], Table]:
    """Or how to make psycopg SLOW."""
    events = []
    row_codec = query.codec
    if issubclass(query.codec, Table):
        row_codec = query.codec.type()
    # ToDo (tribunsky.kir): sqlite raises `AttributeError: __enter__`
    # with conn.cursor() as cur:
    cur = conn.cursor()
    logger.info(f'Executing query: {query.text}')
    cur.execute(query.text)
    logger.info('Validating')
    # ToDo (tribunsky.kir): sqlite returns only tuple with names
    # keys = [col.name for col in cur.description]
    keys = [col[0] for col in cur.description]
    fields = inspect.signature(row_codec).parameters
    fields_match = set(keys) == set(fields)
    if fields_match is False:
        logger.warning("Looks like query doesn't exactly matches model. Please, check.")
    for row in cur:
        row_dict = {key: value for key, value in zip(keys, row)}
        if fields_match is False:
            row_dict = {key: row_dict[key] for key in fields}
        events.append(row_codec(**row_dict))
    if issubclass(query.codec, Table):
        return query.codec(rows=events)
    return events


class SQLReader(Source):

    def __init__(self, queries: List[Query], conn: Callable[[], Connection]) -> None:
        self._conn_factory = conn
        self._queries = queries
        self.patch_annotations({query.codec for query in queries})

    def __call__(self) -> List[Event]:
        conn = self._conn_factory()
        try:
            # with conn:
            rows_or_tables = self._run_queries(conn)
        finally:
            conn.close()
        return rows_or_tables

    def _run_queries(self, conn: Any) -> List[Event]:
        rows_or_tables: List[Event] = []
        for query in self._queries:
            row_or_table = _get_rows_or_table(conn, query)
            if isinstance(row_or_table, list):
                rows_or_tables.extend(row_or_table)
            else:
                rows_or_tables.append(row_or_table)
        return rows_or_tables




from dataclasses import dataclass

from blocks.sqlite import SQLiteReader, SQLiteWriter
from blocks.db.types import Row
from blocks.db.next.sql import Query, Dialects


@dataclass
class TableRow(Row):
    id: int
    name: str
    text: str


def test_read_write(connect) -> None:
    read_query = Query(TableRow, dialect=Dialects.sqlite).SELECT('*').FROM('test_table')
    reader = SQLiteReader([read_query], connect)
    events = reader()
    assert events == []

    insert_query = Query(dialect=Dialects.sqlite).INSERT(TableRow).INTO('test_table')
    writer = SQLiteWriter([insert_query], connect)
    row_original = TableRow(id=1, name='one', text='test')
    writer(row_original)

    [event] = reader()
    assert event == row_original

    update_query = Query(dialect=Dialects.sqlite).UPDATE('test_table').SET(TableRow).WHERE('id')
    writer = SQLiteWriter([update_query], connect)
    row_updated = TableRow(id=1, name='one', text='updated')

    writer(row_updated)

    [event] = reader()
    assert event == row_updated



import time
from typing import Type, Deque, Union, Optional
from collections import deque

from blocks.types import Event, Processor


def _timeout_exceed(timeout: float, ts: float) -> bool:
    return (time.time() - ts) > timeout


class Batcher(Processor):
    """
    Processor that implements batching of events by caching
    them in accumulator until accumulator size reached batch size.
    Then flushes events and clear accumulator. On initialization
    requires event for batching, event for list of events wrapping
    and batch size.

    Example::

      >>> from typing import List

      >>> from blocks import Event
      >>> from blocks.processors import Batcher

      >>> class MyEvent(Event):
      ...     field1: int
      ...     field2: str

      >>> class BatchedEvents(Event):
      ...     events: List[MyEvent]

      >>> batcher = Batcher(MyEvent, BatchedEvents, 2)
    """

    def __init__(
        self,
        input_event: Type[Event],
        batch_event: Type[Event],
        batch_size: int,
    ) -> None:
        self._internal_deque: Deque[Event] = deque()
        self._batch_size = batch_size
        self._input_event = input_event
        self._batch_event = batch_event
        self._batch_name = list(batch_event.__annotations__.keys())[0]
        self.__call__.__annotations__.update({'event': self._input_event, 'return': self._batch_event})

    def __call__(self, event: Event) -> Optional[Event]:
        self._internal_deque.append(event)
        if len(self._internal_deque) == self._batch_size:
            return self._get_batch()
        return None

    def _get_batch(self) -> Event:
        items = list(self._internal_deque)
        self._internal_deque.clear()
        return self._batch_event(**{self._batch_name: items})


class TimeoutedBatcher(Batcher):
    """
    Processor that implements batching of events by caching
    them in accumulator until accumulator size reached batch size.
    Then flushes events and clear accumulator. On initialization
    requires event for batching, event for list of events wrapping,
    batch size and timeout in seconds. Almost identical to Batcher
    with one key difference that TimeoutedBatcher performs early flush
    of accumulator if timeout exceeded. Timeout nullifies when
    batching event received.

    Example::

      >>> from typing import List

      >>> from blocks import Event
      >>> from blocks.processors import TimeoutedBatcher

      >>> class MyEvent(Event):
      ...     field1: int
      ...     field2: str

      >>> class TriggerEvent(Event):
      ...     field1: int

      >>> class BatchedEvents(Event):
      ...     events: List[MyEvent]

      >>> batcher = TimeoutedBatcher(MyEvent, TriggerEvent, BatchedEvents, 2, 1.5)
    """

    def __init__(
        self,
        input_event: Type[Event],
        trigger_event: Type[Event],
        batch_event: Type[Event],
        batch_size: int,
        timeout: float,
    ) -> None:
        self._timeout = timeout
        self._trigger_event = trigger_event
        self._last_assembly_time: Optional[float] = None
        super().__init__(input_event, batch_event, batch_size)
        self.__call__.__annotations__.update({
            'event': Union[self._input_event, self._trigger_event] if trigger_event else input_event,
            'return': batch_event,
        })

    def __call__(self, event: Event) -> Optional[Event]:
        if isinstance(event, self._input_event):
            if self._last_assembly_time is None:
                self._last_assembly_time = time.time()
            batch = super().__call__(event)
            if batch:
                self._last_assembly_time = time.time()
                return batch
        elif (
            isinstance(event, self._trigger_event)
            and self._internal_deque
            and self._last_assembly_time is not None
            and _timeout_exceed(self._timeout, self._last_assembly_time)
        ):
            self._last_assembly_time = time.time()
            return self._get_batch()
        return None



from dataclasses import dataclass
from unittest.mock import MagicMock
from typing import NamedTuple

import pytest
from blocks.runners import Runner, AsyncRunner, is_named_tuple, EventOrEvents, Event


@dataclass
class SimpleEvent(Event):
    value: int = 42


@dataclass
class TerminalEvent(Event):
    ...


class EmptyNamedTuple(NamedTuple):
    ...


class NotEmptyNamedTuple(NamedTuple):
    values: int = 42


events_by_status = (
    (
        None, False
    ),
    (
        SimpleEvent(), False
    ),
    (
        (SimpleEvent(), SimpleEvent()), False
    ),
    (
        [SimpleEvent(), SimpleEvent()], False
    ),
    (
        (EmptyNamedTuple(), EmptyNamedTuple()), False
    ),
    (
        [EmptyNamedTuple(), EmptyNamedTuple()], False
    ),
    (
        (NotEmptyNamedTuple(), NotEmptyNamedTuple()), False
    ),
    (
        [NotEmptyNamedTuple(), NotEmptyNamedTuple()], False
    ),
    (
        (NotEmptyNamedTuple(), EmptyNamedTuple()), False
    ),
    (
        [NotEmptyNamedTuple(), EmptyNamedTuple()], False
    ),
    (
        EmptyNamedTuple(), True
    ),
    (
        NotEmptyNamedTuple(), True
    ),
    (
        TerminalEvent(), False
    )
)

events = [events for events, status in events_by_status]


@pytest.mark.parametrize("events, status", events_by_status)
def test_is_named_tuple(events: EventOrEvents, status: bool):
    assert is_named_tuple(events) == status


@pytest.mark.parametrize("events", events)
def test_runner_append_event(events: EventOrEvents):
    graph = MagicMock()
    graph.contains_async_blocks = False
    graph.count_of_parallel_tasks = 0
    runner = Runner(graph=graph, terminal_event=TerminalEvent)
    runner.stop = MagicMock()
    runner._q = MagicMock()
    runner._append_events(events)
    if isinstance(events, TerminalEvent):
        assert len(runner._q) == 0
        runner._q.appendleft.assert_not_called()
        runner.stop.assert_called()
    elif isinstance(events, (NotEmptyNamedTuple, EmptyNamedTuple)):
        runner._q.appendleft.assert_called_once_with(events)
    elif isinstance(events, (tuple, list)):
        for event in events:
            runner._q.appendleft.assert_any_call(event)
    elif events is None:
        runner._q.appendleft.assert_not_called()
    elif isinstance(events, Event):
        runner._q.appendleft.assert_called_once_with(events)


@pytest.mark.parametrize("events", events)
def test_async_runner_append_event(events: EventOrEvents):
    graph = MagicMock()
    runner = AsyncRunner(graph=graph, terminal_event=TerminalEvent)
    runner.stop = MagicMock()
    runner._q = MagicMock()
    runner._append_events(events)
    if isinstance(events, TerminalEvent):
        assert len(runner._q) == 0
        runner._q.append.assert_not_called()
        runner.stop.assert_called()
    elif isinstance(events, (NotEmptyNamedTuple, EmptyNamedTuple)):
        runner._q.append.assert_called_once_with(events)
    elif isinstance(events, (tuple, list)):
        for event in events:
            runner._q.append.assert_any_call(event)
    elif events is None:
        runner._q.append.assert_not_called()
    elif isinstance(events, Event):
        runner._q.append.assert_called_once_with(events)



from __future__ import annotations

from typing import Any, Type, Union
from datetime import datetime

from schedule import Job

from blocks.types import Event, Source, EventOrEvents


class Scheduler(Source):
    """
    Class represents source of events which are emitted by schedule.

    Therefore, other processors in your program may use these events as time-driven triggers.

    Uses "schedule" library (https://schedule.readthedocs.io/en/stable/)
    under the hood and exposes only subset of it's API.

    Example::

      >>> from blocks import Event
      >>> from blocks.sources.schedule import Scheduler

      >>> class MyEvent(Event):
      ...     pass

      >>> blocks = (Scheduler(MyEvent).every().minute.at(':30'), ...)
    """

    def __init__(self, scheduled_event: Type[Event], *, skip_when_busy: bool = True) -> None:
        self._event = scheduled_event
        self._job = Job(1)
        # FixMe (tribunsky.kir): def fun2(x: tp) -> None: ...  # Error: "tp" is not valid as a type
        self.patch_annotations({scheduled_event})

        self._skip_when_busy = skip_when_busy

    def __call__(self, _: Any = None) -> EventOrEvents:
        if self._skip_when_busy:
            if self._job.should_run:
                return self.__emit_single_event()
        else:
            if not self._job.last_run:
                return self.__emit_single_event()
            if self._job.last_run is None or self._job.period is None:
                raise RuntimeError("Couldn't determine {0} settings, please set-up {0}".format(self.__class__.__name__))
            if datetime.now() - self._job.last_run > self._job.period:
                missed_runs = (datetime.now() - self._job.last_run) // self._job.period
                self._job.last_run = datetime.now()
                return [self._event() for _ in range(missed_runs)]
        return []

    def every(self, interval: int = 1) -> Scheduler:
        self._job.interval = interval
        return self

    @property
    def second(self) -> Scheduler:
        return self.__hide_schedule_api('seconds')

    @property
    def minute(self) -> Scheduler:
        return self.__hide_schedule_api('minutes')

    @property
    def hour(self) -> Scheduler:
        return self.__hide_schedule_api('hours')

    @property
    def seconds(self) -> Scheduler:
        return self.second

    @property
    def minutes(self) -> Scheduler:
        return self.minute

    @property
    def hours(self) -> Scheduler:
        return self.hour

    def at(self, time_str: str) -> Scheduler:
        self._job.at(time_str)
        self._job._schedule_next_run()
        return self

    def __hide_schedule_api(self, unit: str) -> Scheduler:
        self._job.unit = unit
        self._job._schedule_next_run()
        return self

    def __emit_single_event(self) -> EventOrEvents:
        self._job.last_run = datetime.now()
        self._job._schedule_next_run()
        return [self._event()]



import pytest

try:
    from pydantic import BaseModel
except ImportError:
    pytest.skip("skipping pydantic-only tests", allow_module_level=True)

from blocks.redis.serdes import serialize, deserialize


class ManyFieldsModel(BaseModel):
    flag: bool = True
    string: str = 'string'
    pi: float = 3.15
    digit: int = 13
    other_flag: bool = False


def test_ser_des_pydantic() -> None:
    value = ManyFieldsModel()
    serialized = serialize(value.dict())
    deserialized = deserialize(serialized)
    assert value == ManyFieldsModel(**deserialized)



from typing import List, Union, Callable, Any

from blocks import Processor
from blocks.logger import logger
from blocks.db.types import Query, Row, Table

from pyhive.hive import Connection


def _exec_queries(conn: Connection, query_text: str, partition_key: Union[str, None], rows: Table) -> None:
    total_count = len(rows.values)
    with conn.cursor() as cursor:
        for ix, row in enumerate(rows.rows, 1):
            row_dict = row.as_dict
            query = query_text
            query_args: List[Any] = []

            if ' partition ' in query_text.lower():
                if partition_key is None:
                    raise KeyError("Please specify partition key for Query object if using PARTITION in query!")

                partition_value = row_dict.get(partition_key)
                if not partition_value:
                    raise KeyError("Missed partition key column '{0}' in row table. Row columns: {1}".format(
                            partition_key, list(row_dict.keys())))

                query = query.format('{0}=%s'.format(partition_key), '{0}', '{1}')
                query_args.append(partition_value)
                del row_dict[partition_key]

            columns = row_dict.keys()
            query = query.format(
                ', '.join([col for col in columns]),
                ', '.join(['%s' for i in range(len(columns))])
            )
            query_args += row_dict.values()

            logger.info('[{0}/{1}] Executing query: {2} {3}'.format(ix, total_count, query, query_args))
            cursor.execute(query, query_args)
            logger.info('[{0}/{1}] Query execution finished'.format(ix, total_count))


class HiveWriter(Processor):
    """
        Class represents event processor that writes messages to HiveWriter tables
        based on data received from event. On initialization must get list
        of queries in order to perform arbitrary mapping.
        Query arguments which will be pasted from instance of Table:
        - {0} partition key and value, takes from table row, need to pass partition key as argument to Query
        - {1} table columns
        - {2} table values
        Example::

          >>> from dataclasses import dataclass
          >>>
          >>> from blocks import Graph
          >>> from blocks.hive import HiveWriter, Query, Row

          >>> @dataclass
          ... class TableRow(Row):
          ...     partition_col: str
          ...     x: int

          >>> partition_key = 'partition_col'
          >>> queries = [Query('INSERT INTO table PARTITION ({0}) ({1}) VALUES ({2})', TableRow, partition_key)]
          >>> blocks = (HiveWriter(queries), ...)
        """
    def __init__(self, queries: List[Query], connection_factory: Callable[[], Connection]) -> None:
        self._connection_factory = connection_factory
        self._conn = self._connection_factory()
        self._queries = {query.codec: query for query in queries}
        self.__call__.__annotations__.update(
            {
                'event': Union[tuple(query.codec for query in queries)],
            },
        )

    def __call__(self, event: Union[Row, Table]) -> None:
        with self._conn:
            self._run_queries(self._conn, event)

    def _run_queries(self, conn: Connection, event: Union[Row, Table]) -> None:
        query = self._queries[type(event)]
        if isinstance(event, Row):
            table = Table(rows=[event])
        elif isinstance(event, Table):
            table = event
        else:
            raise ValueError('Wrong event type {0}'.format(event))
        _exec_queries(conn, query.text, query.partition_key, table)



import sys
import random
from time import time

from blocks.annotations import get_output_events_type
from blocks.compat import HAS_PYDANTIC

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from typing import Any, Dict, List, Optional, NamedTuple
from dataclasses import dataclass

import pytest

try:
    import wunderkafka
except ImportError:
    pytest.skip("skipping kafka-only tests", allow_module_level=True)

from wunderkafka import AvroConsumer
from confluent_kafka import KafkaError

from blocks.kafka import InputTopic
from blocks.kafka.topics import ConsumerFactory
from blocks.kafka.sources import AnyConsumer, KafkaSource, ConsumerConfig


class Message(Protocol):
    def value(self) -> Dict[str, Any]:
        ...

    def key(self) -> Optional[Dict[str, Any]]:
        ...

    def topic(self) -> Optional[str]:
        ...

    def partition(self) -> Optional[int]:
        ...

    def offset(self) -> Optional[int]:
        ...

    def error(self) -> Optional[KafkaError]:
        ...


class MessageMock(Message):
    def __init__(
        self,
        topic: str,
        value: Dict[str, Any],
        key: Optional[Dict[str, Any]] = None,
        error: Optional[KafkaError] = None,
    ) -> None:
        self._topic = topic
        self._value = value
        self._key = key
        self._error = error

    def value(self) -> Dict[str, Any]:
        return self._value

    def topic(self) -> Optional[str]:
        return self._topic


class ConsumerStub(AnyConsumer):

    def __init__(self, config: ConsumerConfig):
        super().__init__(config)

    def consume(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        ignore_keys: bool = False,
    ) -> List[Message]:
        max_value = 9000
        messages = []
        for _ in range(10):
            msg = {'id': random.randint(-max_value, max_value), 'value': random.uniform(-max_value, max_value)}
            messages.append(MessageMock(topic='test', value=msg))
        return messages


class SignalNT(NamedTuple):
    id: int
    value: float


class SignalO(object):
    id: int
    value: float


class SignalOI(object):
    def __init__(self, id: int, value: float) -> None:
        self.id = id
        self.value = value


@dataclass
class SignalDC:
    id: int
    value: float


class BatchSignal(NamedTuple):
    events: List[SignalNT]


test_types = [
    SignalNT,
    SignalO,
    SignalOI,
    SignalDC,
]

if HAS_PYDANTIC:
    from pydantic import BaseModel


    class SignalP(BaseModel):
        id: int
        value: float


    test_types.append(SignalP)


@pytest.mark.parametrize('cls', test_types)
def test_smoke_event_creation(cls) -> None:
    topics = [InputTopic(name='test', event=cls)]
    source = KafkaSource(topics, ConsumerConfig(group_id='test'), cls=ConsumerStub, ignore_errors=False)

    events = source()
    assert events
    assert all(isinstance(event, cls) for event in events)


def test_smoke_namedtuple_event_topic_override() -> None:
    topics = [InputTopic(name='test', event=SignalNT, consumer=ConsumerFactory(cls=AvroConsumer))]
    with pytest.raises(ValueError):
        KafkaSource(topics, ConsumerConfig(group_id='test'), cls=ConsumerStub)

@pytest.mark.parametrize("event,batch_event,till", [
    (SignalNT, BatchSignal, int(time() * 1000)),
    (SignalNT, BatchSignal, None),
])
def test_batch_annotations(event, batch_event, till) -> None:
    topics = [InputTopic(name='test', event=event, batch_event=batch_event, read_till=till)]
    source = KafkaSource(topics, ConsumerConfig(group_id='test'), cls=ConsumerStub, ignore_errors=False)
    events = get_output_events_type(source)
    assert len(events) == 1
    assert events[0] is BatchSignal



import os.path
from typing import Union
from dataclasses import dataclass

import pytest

from blocks import Graph, source, processor
from blocks.visualization import HAS_NETWORKX, HAS_MATPLOTLIB

if not (HAS_NETWORKX and HAS_MATPLOTLIB):
    pytest.skip("skipping visualization-only tests", allow_module_level=True)

from blocks.types.graph import RenderingKernelType


@dataclass
class KafkaMessage:
    x: int


@dataclass
class MappingWithSize:
    msg: KafkaMessage
    width: int
    height: int


@dataclass
class Meta:
    x: int
    meta: str


@dataclass
class OutputMapping:
    msg: KafkaMessage
    height: int
    width: int
    meta: str


@source
def FastKafkaConsumerMessages() -> KafkaMessage:
    return KafkaMessage(x=1)


@source
def ConsumerMeta() -> Meta:
    return Meta(meta='meta', x=1)


@processor
def MapperSize(e: KafkaMessage) -> MappingWithSize:
    return MappingWithSize(msg=e, width=100, height=200)


@processor
def Predictor(e: Union[MappingWithSize, Meta]) -> OutputMapping:
    print(e)
    return OutputMapping(msg=KafkaMessage(x=1), height=200, width=100, meta='meta')


def test_matplotlib_visualization() -> None:
    blocks = (FastKafkaConsumerMessages(), ConsumerMeta(), MapperSize(), Predictor())
    graph = Graph(blocks)
    graph.save_visualization(rendering_type=RenderingKernelType.matplotlib)
    assert os.path.exists('GRAPH.png')


def test_graphviz_visualization() -> None:
    blocks = (FastKafkaConsumerMessages(), ConsumerMeta(), MapperSize(), Predictor())
    graph = Graph(blocks)
    graph.save_visualization(rendering_type=RenderingKernelType.graphviz)
    assert os.path.exists('DAG.svg')

